/*written by Mattias Heinrich.
Copyright (c) 2015. All rights reserved.
See the LICENSE.txt file in the root folder

contact: heinrich(at)imi.uni-luebeck.de
www.mpheinrich.de

If you use this implementation or parts of it please cite:

"Multi-Atlas Segmentation using Patch-Based Joint Label Fusion with Non-Negative Least Squares Regression."
by Mattias Heinrich, Matthias Wilms, Heinz Handels
First International Workshop on Patch-based Techniques in Medical Imaging, MICCAI 2015

Tested with g++ and clang++ on Mac OS X and Linux Ubuntu, compile with:

g++ NNLSfusion.cpp -O3 std=c++11 -msse4.2 -o nnlsFusion

replace msse4.2 by your current SSE version if needed

This is intended for research purposes only and should not be used in clinical applications (no warranty)!

Example usage (multiple atlases, i.e. more than 2 can be provided)
=============================================================
Usage (required input arguments):
./nnlsFusion -F target.nii.gz -A atlas1.nii.gz atlasn.nii.gz -S seg1.nii.gz segn.nii.gz -O outputname
optional parameters:
-r <radius of patches for fusion (default 4)
-e <epsilon for least-squares Tikhnonov regualisation> (default 0.1)
=============================================================

If you find any bugs or have questions please don’t hesitate to contact me by email.*/

#include <sstream>
#include <cstring>
#include <fstream>
#include <iostream>
#include <vector>
#include <cmath>
#include <map>
#include <queue>
#include <algorithm>
#include <numeric>
#include <functional>
#include <thread>
#include <sys/time.h>
#include <pthread.h>

using namespace std;


struct Image{
    float* values; uint64_t* descriptor;
    int m; int n; int o;
};

struct Options{
    char* file_fixed;
    char* file_output;
    vector<string> files_atlas;
    vector<string> files_seg;
    int radius;
    float epsilon;
};


template <typename TypeFO>
void writeOutput(TypeFO* data,string filestr,int length){
    //opens file for binary-output
    char* filename=new char[filestr.size()+1];
    copy(filestr.begin(),filestr.end(),filename);
    filename[filestr.size()]='\0';
    ofstream ofs1(filename,ios::out|ios::binary);
    ofs1.write(reinterpret_cast<char*>(data),length*sizeof(TypeFO));
    ofs1.close();
}

template <typename TypeF>
vector<TypeF> readFile(char filename[]){
    //opens file for binary-input
    ifstream file(filename,ios::binary|ios::ate);
    vector<TypeF> invalues;
    if(file.is_open()){
        int length=file.tellg()/(sizeof(TypeF));
        file.seekg(0,file.beg); //go to beginning
        char* charptr=new char[length*sizeof(TypeF)];
        file.read((char*)charptr,length*sizeof(TypeF));
        TypeF* floatptr=reinterpret_cast<TypeF*>(charptr);
        for(int i=0;i<length;i++){
            invalues.push_back(floatptr[i]);
        }
        delete charptr;
    }
    file.close();
    return invalues;
}

#include <Eigen/Dense>
#include <Eigen/QR>
#include "nnls.h"

using namespace Eigen;


#include "niftiInOutgz.h"
#include "NNLS3dnorm.h"


Options parseArguments(int ,char* []);
float diceCoef(short* ,short* ,int ,int );

void weightedVoting(short* ,double* ,vector<uint64_t> ,short* ,short* ,double* ,int ,int ,int ,int ,int ,int );

int main(int argc,char* argv[]){
    
    timeval time1,time2;
    
    Options opts=parseArguments(argc,argv);

    
    Image target=readImage(opts.file_fixed);
    int m=target.m; int n=target.n; int o=target.o; uint64_t sz=m*n*o;
    
    //double minval=-1024.0; if(modality[0]=='m'){minval=0.0;cout<<"minval: "<<minval<<"\n";}
    
    double* target_im=new double[sz];
    for(int i=0;i<sz;i++){
        target_im[i]=(double)target.values[i];//max((double)target.values[i],minval);
    }
    
    
    int num_atlas=opts.files_atlas.size();
    
    
    int maxlab=0;
    
    int rad=opts.radius;
    double eps=opts.epsilon;
    
    bool writeResults=true;
    
    
    
    
    double* atlas_im=new double[sz*num_atlas];
    short* atlas_seg=new short[sz*num_atlas];
    for(int a=0;a<num_atlas;a++){
        char* atlasname=new char[200];
        Image atlas=readImage(opts.files_atlas[a]);
        for(int i=0;i<sz;i++){
            atlas_im[i+a*sz]=(double)atlas.values[i];//max(,minval);
        }
        
        delete atlas.values;
        Image atlas2=readImage(opts.files_seg[a]);
        for(int i=0;i<sz;i++){
            short lab1=atlas2.values[i];
            atlas_seg[i+a*sz]=lab1;
        }
        maxlab=max(maxlab,(int)*max_element(atlas2.values,atlas2.values+sz));
        
        delete atlas2.values;
        
    }
    printf("succesfully read %d atlases, maximum label: %d\n",num_atlas,maxlab);
    
    short* majorityVote=new short[sz*(maxlab+1)];
    
    for(int i=0;i<sz*(maxlab+1);i++){
        majorityVote[i]=0;
    }

    for(int a=0;a<num_atlas;a++){
        for(int i=0;i<sz;i++){
            short lab1=atlas_seg[i+a*sz];
            majorityVote[i*(maxlab+1)+lab1]++;
            
        }
    }
    
    short* majority=new short[sz];
    for(int i=0;i<sz;i++){
        majority[i]=0;
    }
    double* unanimous=new double[sz];
    
    for(int i=0;i<sz;i++){
        majority[i]=max_element(majorityVote+i*(maxlab+1),majorityVote+(i+1)*(maxlab+1))-(majorityVote+i*(maxlab+1));
        if(*max_element(majorityVote+i*(maxlab+1),majorityVote+(i+1)*(maxlab+1))==num_atlas){
            unanimous[i]=1;
        }
        else{
            unanimous[i]=0;
        }
    }
    
    char percent[3];
    strcpy(percent,"%");
    printf("majority voting done, %2.3f %s unanimous\n",accumulate(unanimous,unanimous+sz,0.0)/(double)sz*100.0,percent);
    
    /*
     //eval
     sprintf(targetname,"_seg.nii.gz",modality,targetnum);
     
     Image target_seg1=readImage(targetname);
     short* target_seg=new short[sz];
     for(int i=0;i<sz;i++){
     short lab1=(target_seg1.values[i]>=8)?0:target_seg1.values[i];
     target_seg[i]=lab1;
     }
     
     float d_maj=diceCoef(target_seg,majority,sz,maxlab);
     */
    double* ones=new double[sz];
    double* ones1=new double[sz];
    double* temp1=new double[sz];
    double* temp2=new double[sz];
    
    double* mask_dilated=new double[sz];
    for(int i=0;i<sz;i++){
        ones1[i]=1.0;
        ones[i]=1.0;
        mask_dilated[i]=1.0-unanimous[i];
    }
    boxfilter3dref(ones,temp1,temp2,ones1,rad,m,n,o);
    for(int i=0;i<sz;i++){
        ones[i]=1.0/ones[i]; //factor for meanfilter
    }
    boxfilter3dref(mask_dilated,temp1,temp2,ones1,rad,m,n,o);
    int mask_num=sz-count(mask_dilated,mask_dilated+sz,0);
    cout<<"count_if: "<<mask_num<<"\n";
    
    //write out majority vote
    Image seg_out; seg_out.m=m; seg_out.n=n; seg_out.o=o;
    seg_out.values=new float[sz];
    for(int i=0;i<sz;i++){
        seg_out.values[i]=majority[i];
    }
    vector<float> voxelsize(3,1.0f);
    //vector<float> v_float(v_int.begin(), v_int.end());
    char* outname=new char[200];
    sprintf(outname,"%s_majority.nii.gz",opts.file_output);
    if(writeResults){
        writeImage(outname,seg_out,voxelsize);
    }
    
    double* w_nnls=new double[mask_num*num_atlas];
    double* w_lsq=new double[mask_num*num_atlas];
    double* w_ncc=new double[mask_num*num_atlas];
    
    pair<double,double> timePair={0,0};
    vector<uint64_t> inverse_ind=jointNNLSfilter(w_nnls,w_lsq,w_ncc,target_im,atlas_im,mask_dilated,m,n,o,num_atlas,rad,eps,timePair);
    
    //write NNLS
    short* w_seg=new short[sz];
    weightedVoting(w_seg,w_nnls,inverse_ind,atlas_seg,majority,unanimous,m,n,o,num_atlas,rad,maxlab);
    //float dc=diceCoef(target_seg,w_seg,sz,maxlab);
    for(int i=0;i<sz;i++){
        seg_out.values[i]=w_seg[i];
    }
    sprintf(outname,"%s_nnls.nii.gz",opts.file_output);
    if(writeResults){
        writeImage(outname,seg_out,voxelsize);
    }
    
    //write LSQ
    weightedVoting(w_seg,w_lsq,inverse_ind,atlas_seg,majority,unanimous,m,n,o,num_atlas,rad,maxlab);
    //dc=diceCoef(target_seg,w_seg,sz,maxlab);
    for(int i=0;i<sz;i++){
        seg_out.values[i]=w_seg[i];
    }
    sprintf(outname,"%s_lsq.nii.gz",opts.file_output);
    if(writeResults){
        writeImage(outname,seg_out,voxelsize);
    }
    //write NCC
    weightedVoting(w_seg,w_ncc,inverse_ind,atlas_seg,majority,unanimous,m,n,o,num_atlas,rad,maxlab);
    //dc=diceCoef(target_seg,w_seg,sz,maxlab);
    for(int i=0;i<sz;i++){
        seg_out.values[i]=w_seg[i];
    }
    sprintf(outname,"%s_ncc.nii.gz",opts.file_output);
    if(writeResults){
        writeImage(outname,seg_out,voxelsize);
    }
    
    
    
    
    
    delete w_nnls; delete w_lsq; delete w_ncc;
    delete temp1; delete temp2; delete ones; delete ones1; delete w_seg; delete majority; //delete target_seg;
    delete unanimous; delete majorityVote;
    delete atlas_im; delete atlas_seg; delete target_im;
    

    
    
    return 0;
}

float diceCoef(short* segment1,short* segment2,int sz,int max_label){
    
    auto printall=[](vector<float> arr){for(auto &x : arr){printf("%2.3f ",x);}cout<<'\n';};
    
    
    vector<float> dice1;
    
    bool* countvol=new bool[sz];
    //int max_label=8;//*max_element(segment1,segment1+sz)+1;
    for(int label=1;label<=max_label;label++){ //for all labels except background
        
        
        for(int i=0;i<sz;i++)
            countvol[i]=segment1[i]==label&segment2[i]==label;
        float unionval=count(countvol,countvol+sz,true);
        for(int i=0;i<sz;i++)
            countvol[i]=segment1[i]==label;
        float card1=count(countvol,countvol+sz,true);
        for(int i=0;i<sz;i++)
            countvol[i]=segment2[i]==label;
        float card2=count(countvol,countvol+sz,true);
        if(card1+card2>0){
            dice1.push_back(2.0f*unionval/(card1+card2));
        }
        else{
            dice1.push_back(0.0f);
        }
        
        
    }
    printall(dice1);
    float meandice=accumulate(dice1.begin(),dice1.end(),0.0f)/dice1.size();
    
    printf("Mean DICE: %2.4f\n",meandice);
    return meandice;
}
    
Options parseArguments(int argc,char* argv[]){
    
    Options opts;
    
    if(argc<5||argv[1][1]=='h'){
        cout<<"=============================================================\n";
        cout<<"Usage (required input arguments):\n";
        cout<<"./nnlsFusion -F target.nii.gz -A atlas1.nii.gz atlasn.nii.gz -S seg1.nii.gz segn.nii.gz -O outputname\n";
        cout<<"optional parameters:\n";
        cout<<" -r <radius of patches for fusion (default 4)\n";
        cout<<" -e <epsilon for least-squares Tikhnonov regualisation> (default 0.1)\n";
        //cout<<" -G <ground truth segmentation of target for evaluation>\n";
        cout<<"=============================================================\n";
        exit(-1);
    }
    
    
    typedef pair<char,int> val;
    map<char,int> argin;
    argin.insert(val('F',0));
    argin.insert(val('A',1));
    argin.insert(val('S',2));
    argin.insert(val('O',3));
    argin.insert(val('r',4));
    argin.insert(val('e',5));
    
    // parsing the input
    int requiredArgs=0;
    opts.file_fixed=new char[200];
    opts.file_output=new char[200];
    opts.epsilon=0.1;
    opts.radius=4;
    
    char* filename=new char[200];

    int num;
   
    
    for(int k=1;k<argc;k++){
        if(argv[k][0]=='-'){
            if(argin.find(argv[k][1])==argin.end()){
                cout<<"Invalid option: "<<argv[k]<<" use -h for help\n";
            }
            switch(argin[argv[k][1]]){
                case 0: //F
                    sprintf(opts.file_fixed,"%s",argv[k+1]);
                    requiredArgs++;
                    break;
                case 1: //A
                    sprintf(filename,"%s",argv[k+1]);
                    opts.files_atlas.push_back(filename);
                    while(k+2<argc&&argv[k+2][0]!='-'){
                        k++;
                        sprintf(filename,"%s",argv[k+1]);
                        opts.files_atlas.push_back(filename);
                    }
                    requiredArgs++;
                    break;
                case 2: //S
                    sprintf(filename,"%s",argv[k+1]);
                    opts.files_seg.push_back(filename);
                    while(k+2<argc&&argv[k+2][0]!='-'){
                        k++;
                        sprintf(filename,"%s",argv[k+1]);
                        opts.files_seg.push_back(filename);
                    }
                    requiredArgs++;
                    break;
                case 3: //O
                    sprintf(opts.file_output,"%s",argv[k+1]);
                    requiredArgs++;
                    break;
                case 4: //r
                    opts.radius=atoi(argv[k+1]);
                    break;
                case 5: //e
                    opts.epsilon=atof(argv[k+1]);
                    break;
               
             
                default:
                    cout<<"Invalid option: "<<argv[k]<<" use -h for help\n";
                    break;
            }
        }
    }
    if(opts.files_seg.size()==0|opts.files_seg.size()!=opts.files_atlas.size()){
        cout<<"Same number of atlas ("<<opts.files_atlas.size()<<") and segmentation ("<<opts.files_seg.size()<<") files must be provided.\n";
        exit(-1);

    }
    
    if(requiredArgs!=4){
        cout<<"Missing argmuents (only "<<requiredArgs<<" provided) use -h for help.\n";
        exit(-1);
    }
    
    
    return opts;
}


void weightedVoting(short* weighted_seg,double* weights_sparse,vector<uint64_t> inverse_ind,short* atlas_seg,short* majority,double* unanimous,int m,int n,int o,int num_atlas,int rad,int maxlab){
    int mask_num=inverse_ind.size();
    
    uint64_t sz=m*n*o;
    
    double* ones=new double[sz];
    double* temp1=new double[sz];
    double* temp2=new double[sz];
    
    for(int i=0;i<sz;i++){
        ones[i]=1.0;
    }
    
    
    double* weights=new double[sz*num_atlas];
    for(uint64_t i=0;i<sz*num_atlas;i++){
        weights[i]=0.0;
    }
    for(int l=0;l<num_atlas;l++){
        for(int i=0;i<mask_num;i++){
            weights[inverse_ind[i]+l*sz]=weights_sparse[i+l*mask_num]; //write in actual image dimensions
        }
        boxfilter3dref(weights+l*sz,temp1,temp2,ones,rad,m,n,o);
    }
    
    
    //weighted voting
    for(int i=0;i<sz;i++){
        if(unanimous[i]==0){
            double sum_weight=0.0;
            for(int l=0;l<num_atlas;l++){
                sum_weight+=weights[i+l*sz];
            }
            vector<double> votes(maxlab+1,0.0);
            for(int l=0;l<num_atlas;l++){
                short lab1=atlas_seg[i+l*sz];
                votes[lab1]+=weights[i+l*sz]/sum_weight;
            }
            weighted_seg[i]=max_element(votes.begin(),votes.end())-votes.begin();
        }
        else{
            weighted_seg[i]=majority[i];
        }
    }
    delete weights; delete ones; delete temp1; delete temp2;
}

